/* Copyright(C) 2004 Brazil

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  */

#include <ruby.h>
#include "senna.h"

static VALUE mSenna, cIndex, cRecords;

typedef struct {
  sen_index *i;
} index_t;

typedef struct {
  sen_records *r;
} records_t;

static void
index_free(index_t *i)
{
  sen_index_close(i->i);
  free(i);
}

static void
records_free(records_t *r)
{
  sen_records_close(r->r);
  free(r);
}

static VALUE
index_sel(VALUE self, VALUE key)
{
  VALUE res;
  index_t *i;
  sen_records *r;
  records_t *rp;
  Check_Type(key, T_STRING);
  Data_Get_Struct(self, index_t, i);
  r = sen_index_sel(i->i, STR2CSTR(key));
  if (!r) { return Qnil; }
  res = Data_Make_Struct(cRecords, records_t, 0, records_free, rp);
  if (!rp) {
    rb_raise(rb_eException, "malloc failed");
    return Qnil;
  }
  rp->r = r;
  return res;
}

static VALUE
index_upd(VALUE self, VALUE key, VALUE oldvalue, VALUE newvalue)
{
  index_t *i;
  sen_rc rc;
  char *ov, *nv;
  Check_Type(key, T_STRING);
  if (NIL_P(oldvalue)) {
    ov = NULL;
  } else {
    Check_Type(oldvalue, T_STRING);
    ov = STR2CSTR(oldvalue);
  }
  if (NIL_P(newvalue)) {
    nv = NULL;
  } else {
    Check_Type(newvalue, T_STRING);
    nv = STR2CSTR(newvalue);
  }
  Data_Get_Struct(self, index_t, i);
  rc = sen_index_upd(i->i, STR2CSTR(key), ov, nv);
  return rc ? Qfalse : Qtrue;
}

static VALUE
index_new(int argc, VALUE *argv, VALUE klass)
{
  index_t *i;
  char *path;
  VALUE self;
  int flags = 0;
  if (argc < 1) {
    rb_raise(rb_eArgError, "invalid argument");
    return Qnil;
  }
  Check_Type(argv[0], T_STRING);
  path = STR2CSTR(argv[0]);
  if (strlen(path) > PATH_MAX - 4) {
    rb_raise(rb_eArgError, "invalid argument (too long path \"%s\"", path);
    return Qnil;
  }
  if (argc > 1) { flags = NUM2INT(argv[1]); }
  self = Data_Make_Struct(klass, index_t, 0, index_free, i);
  if (!i) {
    rb_raise(rb_eException, "malloc failed");
    return Qnil;
  }
  if (!(i->i = sen_index_open(path))) {
    if (!(i->i = sen_index_create(path, 0, SEN_INDEX_NORMALIZE, 0, sen_enc_default))) {
      rb_raise(rb_eException, "malloc failed");
      return Qnil;
    }
  }
  rb_obj_call_init(self, 0, NULL);
  return self;
}

static VALUE
index_check(VALUE self)
{
  index_t *i;
  Data_Get_Struct(self, index_t, i);
  sen_index_check(i->i);
  return Qnil;
}

static VALUE
records_nhits(VALUE self)
{
  records_t *r;
  Data_Get_Struct(self, records_t, r);
  return INT2NUM(sen_records_nhits(r->r));
}

static VALUE
records_rewind(VALUE self)
{
  records_t *r;
  Data_Get_Struct(self, records_t, r);
  return sen_records_rewind(r->r) ? Qfalse : Qtrue;
}

static VALUE
records_curr_score(VALUE self)
{
  records_t *r;
  Data_Get_Struct(self, records_t, r);
  return INT2NUM(sen_records_curr_score(r->r));
}

static VALUE
records_find(VALUE self, VALUE key)
{
  records_t *r;
  Check_Type(key, T_STRING);
  Data_Get_Struct(self, records_t, r);
  return INT2NUM(sen_records_find(r->r, STR2CSTR(key)));
}

static VALUE
records_next(VALUE self)
{
  records_t *r;
  const char *res;
  Data_Get_Struct(self, records_t, r);
  res = sen_records_next(r->r);
  return res ? rb_str_new2(res) : Qnil;
}

void
Init_senna(void)
{
  sen_init();
  mSenna = rb_define_module("Senna");
  cIndex = rb_define_class_under(mSenna, "Index", rb_cObject);
  cRecords = rb_define_class_under(mSenna, "Records", rb_cObject);
  rb_define_singleton_method(cIndex, "new", index_new, -1);
  rb_define_method(cIndex, "upd", index_upd, 3);
  rb_define_method(cIndex, "sel", index_sel, 1);
  rb_define_method(cIndex, "check", index_check, 0);
  rb_define_method(cRecords, "nhits", records_nhits, 0);
  rb_define_method(cRecords, "rewind", records_rewind, 0);
  rb_define_method(cRecords, "find", records_find, 1);
  rb_define_method(cRecords, "next", records_next, 0);
  rb_define_method(cRecords, "curr_score", records_curr_score, 0);
}
