/* Copyright(C) 2004 Brazil

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <string.h>
#include "senna_in.h"
#include "str.h"
#include "set.h"

static sen_set *prefix = NULL;
static sen_set *suffix = NULL;

#define N_PREFIX 2048
#define N_SUFFIX 0

#define PREFIX_PATH SENNA_HOME "/prefix"
#define SUFFIX_PATH SENNA_HOME "/suffix"

inline static void
prefix_init(void)
{
  int i, *ip;
  FILE *fp;
  char buffer[4];
  prefix = sen_set_open(2, sizeof(int), 0);
  if ((fp = fopen(PREFIX_PATH, "r"))) {
    for (i = 0; i < N_PREFIX; i++) {
      if (!fgets(buffer, 4, fp)) { break; }
      sen_set_get(prefix, buffer, (void **)&ip);
      *ip = i;
    }
    fclose(fp);
  }
}

inline static void
suffix_init(void)
{
  int i;
  FILE *fp;
  char buffer[4];
  suffix = sen_set_open(2, 0, 0);
  if ((fp = fopen(SUFFIX_PATH, "r"))) {
    for (i = N_SUFFIX; i; i--) {
      if (!fgets(buffer, 4, fp)) { break; }
      sen_set_get(suffix, buffer, NULL);
    }
    fclose(fp);
  }
}

int
sen_str_get_prefix_order(const char *str)
{
  int *ip;
  if (!prefix) { prefix_init(); }
  if (sen_set_at(prefix, str, (void **)&ip)) {
    return *ip;
  } else {
    return -1;
  }
}

inline static int
char_type(unsigned char *str, int *length)
{
  unsigned char c = *str;

  if (!suffix) { suffix_init(); }

  if (c & 0x80) {
    if (!str[1]) {
      *length = 1;
      return sen_str_others;
    }
    *length = 2;

    if (sen_set_at(suffix, str, NULL)) {
      return sen_str_symbol;
    }

    switch (c >> 4) {
    case 0x08 :
      return (c == 0x8e) ? sen_str_katakana : sen_str_others;
    case 0x09 :
      return sen_str_others;
    case 0x0a :
      switch (c & 0x0f) {
      case 1 :
	switch (str[1]) {
	case 0xbc :
	  return sen_str_katakana;
	case 0xb9 :
	  return sen_str_kanji;
	default :
	  return sen_str_symbol;
	}
      case 2 :
	return sen_str_symbol;
      case 3 :
	return (str[1] <= 0x39) ? sen_str_digit : sen_str_alpha;
      case 4 :
	return sen_str_hiragana;
      case 5 :
	return sen_str_katakana;
      case 6 :
      case 7 :
      case 8 :
	return sen_str_symbol;
      default :
	return sen_str_others;
      }
    default :
      return sen_str_kanji;
    }
  } else {
    *length = 1;
    switch (c >> 4) {
    case 2 :
      return sen_str_symbol;
    case 3 :
      return (c <= 0x39) ? sen_str_digit : sen_str_symbol;
    case 4 :
      return (c == 0x40) ? sen_str_symbol : sen_str_alpha;
    case 5 :
      return (c <= 0x5a) ? sen_str_alpha : sen_str_symbol;
    case 6 :
      return (c == 0x60) ? sen_str_symbol : sen_str_alpha;
    case 7 :
      return (c <= 0x7a) ? sen_str_alpha : (c == 0x7f ? sen_str_others : sen_str_symbol);
    default :
      return sen_str_others;
    }
  }
}

inline static char *
get_substr(char *str, sen_substr *res)
{
  int len, ctype = char_type(str, &len);
  if (ctype == sen_str_symbol) {
    res->ctype = ctype;
    res->start = str;
    res->length = len;
    return str + len;
  } else {
    char *p;
    for (p = str + len; *p; p += len) {
      if (char_type(p, &len) != ctype) { break; }
    }
    res->ctype = ctype;
    res->start = str;
    res->length = p - str;
    return p;
  }
}

int
sen_str_split(char *str, sen_substr *res, int max, char **rest)
{
  int i;
  sen_substr *r;
  char *p = str;
  if (!p) { return 0; }
  for (r = res, i = 0; i < max; r++, i++) {
    if (!*p) { break; }
    p = get_substr(p, r);
  }
  if (rest) { *rest = p; }
  return i;
}

struct _sen_str_splitter {
  char *next;
  char backup; 
};

sen_str_splitter *
sen_str_splitter_open(char *str)
{
  sen_str_splitter *s;
  if (!(s = SEN_MALLOC(sizeof(sen_str_splitter)))) { return NULL; }
  s->next = str;
  s->backup = *str;
  return s;
}

char *
sen_str_splitter_next(sen_str_splitter *s, int *ctype)
{
  int c, l;
  char *p, *q;
  p = s->next;
  *p = s->backup;
  if (!*p) { return NULL; }
  c = char_type(p, &l);
  q = p + l;
  if (c != sen_str_symbol) {
    for (; *q; q += l) {
      if (char_type(q, &l) != c) { break; }
    }
  }
  s->next = q;
  s->backup = *q;
  *q = '\0';
  if (ctype) { *ctype = c; }
  return p;
}

sen_rc
sen_str_splitter_close(sen_str_splitter *s)
{
  if (!s) { return sen_invalid_argument; }
  *s->next = s->backup;
  SEN_FREE(s);
  return sen_success;
}

#define BUFSIZE 65536

#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <sys/time.h>

void
sen_log(char *fmt, ...)
{
  static FILE *fp = NULL;
  va_list argp;
  char buffer[BUFSIZE];
  va_start(argp, fmt);
  vsnprintf(buffer, BUFSIZE - 1, fmt, argp);
  va_end(argp);
  buffer[BUFSIZE - 1] = '\0';
  if (!fp) { fp = fopen(SENNA_HOME "/log/senna.log", "a"); }
  if (fp) {
    struct timeval tv;
    struct tm tm;
    gettimeofday(&tv, NULL);
    localtime_r((time_t *)&tv.tv_sec, &tm);
    fprintf(fp, "%02d/%02d:%02d:%02d:%02d.%06d|%s\n",
	    tm.tm_mon + 1, tm.tm_mday,
	    tm.tm_hour, tm.tm_min, tm.tm_sec, (int) tv.tv_usec, buffer);
    fflush(fp);
  }
}

static unsigned char symbol[] = {
  ',', '.', 0, ':', ';', '?', '!', 0, 0, 0, '`', 0, '^', '~', '_', 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, '-', '-', '/', '\\', 0, 0, '|', 0, 0, 0, '\'', 0,
  '"', '(', ')', 0, 0, '[', ']', '{', '}', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  '+', '-', 0, 0, 0, '=', 0, '<', '>', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  '$', 0, 0, '%', '#', '&', '*', '@', 0, 0, 0, 0, 0, 0, 0, 0
};

static uint16_t hankana[] = {
  0xa1a1, 0xa1a3, 0xa1d6, 0xa1d7, 0xa1a2, 0xa1a6, 0xa5f2, 0xa5a1, 0xa5a3,
  0xa5a5, 0xa5a7, 0xa5a9, 0xa5e3, 0xa5e5, 0xa5e7, 0xa5c3, 0xa1bc, 0xa5a2,
  0xa5a4, 0xa5a6, 0xa5a8, 0xa5aa, 0xa5ab, 0xa5ad, 0xa5af, 0xa5b1, 0xa5b3,
  0xa5b5, 0xa5b7, 0xa5b9, 0xa5bb, 0xa5bd, 0xa5bf, 0xa5c1, 0xa5c4, 0xa5c6,
  0xa5c8, 0xa5ca, 0xa5cb, 0xa5cc, 0xa5cd, 0xa5ce, 0xa5cf, 0xa5d2, 0xa5d5,
  0xa5d8, 0xa5db, 0xa5de, 0xa5df, 0xa5e0, 0xa5e1, 0xa5e2, 0xa5e4, 0xa5e6,
  0xa5e8, 0xa5e9, 0xa5ea, 0xa5eb, 0xa5ec, 0xa5ed, 0xa5ef, 0xa5f3, 0xa1ab,
  0xa1eb
};

static unsigned char dakuten[] = {
  0xf4, 0, 0, 0, 0, 0xac, 0, 0xae, 0, 0xb0, 0, 0xb2, 0, 0xb4, 0, 0xb6, 0,
  0xb8, 0, 0xba, 0, 0xbc, 0, 0xbe, 0, 0xc0, 0, 0xc2, 0, 0, 0xc5, 0, 0xc7,
  0, 0xc9, 0, 0, 0, 0, 0, 0, 0xd0, 0, 0, 0xd3, 0, 0, 0xd6, 0, 0, 0xd9, 0,
  0, 0xdc
};

static unsigned char handaku[] = {
  0xd1, 0, 0, 0xd4, 0, 0, 0xd7, 0, 0, 0xda, 0, 0, 0xdd
};

inline static int
normalize_euc(const char *str, char *norm, uint_least8_t *ctypes)
{
  uint_least8_t *cp = ctypes, ctype;
  const unsigned char *s;
  unsigned char *d, *d0 = (unsigned char *) norm, b;
  for (s = (unsigned char *) str, d = d0; *s; s++) {
    if ((*s & 0x80)) {
      if ((*(s + 1) & 0x80)) {
	unsigned char c1 = *s++, c2 = *s, c3 = 0;
	switch (c1 >> 4) {
	case 0x08 :
	  if (c1 == 0x8e && 0xa0 <= c2 && c2 <= 0xdf) {
	    uint16_t c = hankana[c2 - 0xa0];
	    switch (c) {
	    case 0xa1ab :
	      if (d > d0 + 1 && d[-2] == 0xa5
		  && 0xa6 <= d[-1] && d[-1] <= 0xdb && (b = dakuten[d[-1] - 0xa6])) {
		*(d - 1) = b;
		continue;
	      } else {
		*d++ = c >> 8; *d = c & 0xff;
	      }
	      break;
	    case 0xa1eb :
	      if (d > d0 + 1 && d[-2] == 0xa5
		  && 0xcf <= d[-1] && d[-1] <= 0xdb && (b = handaku[d[-1] - 0xcf])) {
		*(d - 1) = b;
		continue;
	      } else {
		*d++ = c >> 8; *d = c & 0xff;
	      }
	      break;
	    default :
	      *d++ = c >> 8; *d = c & 0xff;
	      break;
	    }
	    ctype = sen_str_katakana;
	  } else {
	    *d++ = c1; *d = c2;
	    ctype = sen_str_others;
	  }
	  break;
	case 0x09 :
	  *d++ = c1; *d = c2;
	  ctype = sen_str_others;
	  break;
	case 0x0a :
	  switch (c1 & 0x0f) {
	  case 1 :
	    switch (c2) {
	    case 0xbc :
	      *d++ = c1; *d = c2;
	      ctype = sen_str_katakana;
	      break;
	    case 0xb9 :
	      *d++ = c1; *d = c2;
	      ctype = sen_str_kanji;
	      break;
	    default :
	      if (c2 == 0xa1) {
		if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
		continue;
		/* skip white space */
	      } else if (c2 >= 0xa4 && (c3 = symbol[c2 - 0xa4])) {
		*d = c3;
		ctype = sen_str_symbol;
	      } else {
		*d++ = c1; *d = c2;
		ctype = sen_str_others;
	      }
	      break;
	    }
	    break;
	  case 2 :
	    *d++ = c1; *d = c2;
	    ctype = sen_str_symbol;
	    break;
	  case 3 :
	    c3 = c2 - 0x80;
	    if ('a' <= c3 && c3 <= 'z') {
	      ctype = sen_str_alpha;
	      *d = c3;
	    } else if ('A' <= c3 && c3 <= 'Z') {
	      ctype = sen_str_alpha;
	      *d = c3 + 0x20;
	    } else if ('0' <= c3 && c3 <= '9') {
	      ctype = sen_str_digit;
	      *d = c3;
	    } else {
	      ctype = sen_str_others;
	      *d++ = c1; *d = c2;
	    }
	    break;
	  case 4 :
	    *d++ = c1; *d = c2;
	    ctype = sen_str_hiragana;
	    break;
	  case 5 :
	    *d++ = c1; *d = c2;
	    ctype = sen_str_katakana;
	    break;
	  case 6 :
	  case 7 :
	  case 8 :
	    *d++ = c1; *d = c2;
	    ctype = sen_str_symbol;
	    break;
	  default :
	    *d++ = c1; *d = c2;
	    ctype = sen_str_others;
	    break;
	  }
	  break;
	default :
	  *d++ = c1; *d = c2;
	  ctype = sen_str_kanji;
	  break;
	}
      } else {
	/* skip invalid character */
	continue;
      }
    } else {
      unsigned char c = *s;
      switch (c >> 4) {
      case 0 :
      case 1 :
	/* skip unprintable ascii */
	if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
	continue;
      case 2 :
	if (c == 0x20) {
	  /* skip white space */
	  if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
	  continue;
	} else {
	  *d = c;
	}
	ctype = sen_str_symbol;
	break;
      case 3 :
	*d = c;
	ctype = (c <= 0x39) ? sen_str_digit : sen_str_symbol;
	break;
      case 4 :
	*d = ('A' <= c) ? c + 0x20 : c;
	ctype = (c == 0x40) ? sen_str_symbol : sen_str_alpha;
	break;
      case 5 :
	*d = (c <= 'Z') ? c + 0x20 : c;
	ctype = (c <= 0x5a) ? sen_str_alpha : sen_str_symbol;
	break;
      case 6 :
	*d = c;
	ctype = (c == 0x60) ? sen_str_symbol : sen_str_alpha;
	break;
      case 7 :
	*d = c;
	ctype = (c <= 0x7a) ? sen_str_alpha : (c == 0x7f ? sen_str_others : sen_str_symbol);
	break;
      default :
	*d = c;
	ctype = sen_str_others;
	break;
      }
    }
    // printf("%c%c %x %x  -> %c %x  %x\n", *(s - 1), *(s), *(s - 1), *(s), *d, *d, ctype);
    *cp++ = ctype;
    d++;
  }
  *cp = sen_str_null;
  *d = '\0';
  return cp - ctypes;
}

inline static int
normalize_utf8(const char *str, char *norm, uint_least8_t *ctypes)
{
  // todo;
  return 0;
}

inline static int
normalize_sjis(const char *str, char *norm, uint_least8_t *ctypes)
{
  // todo;
  return 0;
}

inline static int
normalize_none(const char *str, char *norm, uint_least8_t *ctypes)
{
  // todo;
  return 0;
}

sen_nstr *
sen_nstr_open(const char *str, sen_encoding encoding)
{
  size_t size;
  sen_nstr *r;
  if (!str || !(size = strlen(str))) { return NULL; }
  if (!(r = SEN_MALLOC(sizeof(sen_nstr)))) { return NULL; }
  r->orig = str;
  if (!(r->norm = SEN_MALLOC(size * 2 + 1))) {
    SEN_FREE(r);
    return NULL;
  }
  if (!(r->ctypes = SEN_MALLOC(size + 1))) {
    SEN_FREE(r->norm);
    SEN_FREE(r);
    return NULL;
  }
  switch (encoding) {
  case sen_enc_euc_jp :
    r->length = normalize_euc(str, r->norm, r->ctypes);
    break;
  case sen_enc_utf8 :
    r->length = normalize_utf8(str, r->norm, r->ctypes);
    break;
  case sen_enc_sjis :
    r->length = normalize_sjis(str, r->norm, r->ctypes);
    break;
  default :
    r->length = normalize_none(str, r->norm, r->ctypes);
    break;
  }
  /*
  {
    uint_least8_t *ip;
    for (ip = r->ctypes; *ip; ip++) {
      switch(*ip & 0x7f) {
      case sen_str_alpha : putchar('a'); break;
      case sen_str_digit : putchar('n'); break;
      case sen_str_symbol : putchar('s'); break;
      case sen_str_hiragana : putchar('h'); break;
      case sen_str_katakana : putchar('k'); break;
      case sen_str_kanji : putchar('K'); break;
      case sen_str_others : putchar('o'); break;
      default : putchar('e'); break;	
      }
      if (SEN_NSTR_ISBLANK(*ip)) { putchar(' '); }
    }
    putchar('\n');
  }
  */
  return r;
}

sen_rc
sen_nstr_close(sen_nstr *nstr)
{
  if (!nstr) { return sen_invalid_argument; }
  SEN_FREE(nstr->norm);
  SEN_FREE(nstr->ctypes);
  SEN_FREE(nstr);
  return sen_success;
}

const char * sen_enc_string[] = {
  "default",
  "none",
  "euc_jp",
  "utf8",
  "sjis"
};

size_t
sen_str_charlen(const char *str, sen_encoding encoding)
{
  size_t size;
  const unsigned char *p = (const unsigned char *) str;
  if (!*p) { return 0; }
  switch (encoding) {
  case sen_enc_euc_jp :
    size = ((*p & 0x80) && (*(p + 1))) ? 2 : 1;
    break;
  case sen_enc_utf8 :
    if (*p & 0x80) {
      int b, w; 
      for (b = 0x40, w = 0; b && (*p & b); b >>= 1, w++);
      if (!w) { sen_log("invalid utf8 string"); }
      for (size = 1; w--; size++) {
	if (!*++p || (*p & 0xc0) != 0x80) {
	  sen_log("invalid utf8 string");
	  break;
	}
      }
    } else {
      size = 1;
    }
    break;
  case sen_enc_sjis :
    if (*p & 0x80) {
      if ((0xa1 <= *p && *p <= 0xdf) || !(*(p + 1))) {
	size = 1;
      } else {
	size = 2;
      }
    } else {
      size = 1;
    }
    break;
  default :
    size = 1;
    break;
  }
  return size;
}

size_t
sen_str_len(const char *str, sen_encoding encoding, const char **last)
{
  size_t len;
  const char *p = NULL;
  for (len = 0; *str; len++) {
    p = str;
    str += sen_str_charlen(str, encoding);
  }
  if (last) { *last = p; }
  return len;
}
